#
# T3D exporter for Google SketchUp
# exporder.rb -- extension core
# Created by Ferry Timmers
# June 2010
#

#------------------------------------------------------------------------------#

require 'sketchup.rb'
require 't3d/t3d.rb'

#------------------------------------------------------------------------------#
# Semi globals

@progress = nil
@thread = nil

#------------------------------------------------------------------------------#

UI.menu('PlugIns').add_item('Export to T3D...') { show_dialog }

#------------------------------------------------------------------------------#

def show_dialog
	# Fudge factor: (w+50, h+50)
	dlg = UI::WebDialog.new 'T3D exporter', false, 'T3DExporter',
		550, 350, 200, 200, false
	
	dlg.set_size 550, 350
	
	dlg.set_file Sketchup.find_support_file('dialog.html', 'Plugins/t3d/')
	
	#-------------------------------------
	
	dlg.add_action_callback('choose_file') {|d,a| choose_file d}
	
	#-------------------------------------
	
	dlg.add_action_callback('export_file') do |d,a|
		
		if not @thread.nil? and @thread.alive?
			@thread.kill
			@progress.status 'Aborted'
			@progress.progress 0
			@progress.abort false
			break
		end
		
		@progress = Progress.new(d);
		
		@thread = Thread.new { export d.get_element_value('filename') }
		
		progression = UI.start_timer 0.1, true do
			100.times { Thread.pass }
			UI.stop_timer progression unless @thread.alive?
			@progress.progress
		end
	end
	
	#-------------------------------------
	
	dlg.add_action_callback('show') do |d,a|
		model = Sketchup.active_model
		model = model.selection.empty? ? model.entities : model.selection
		
		case a
			when 'all'
				show_culprits(model) {true}
			
			when 'errors'
				show_culprits(model) do |e|
					defined? e.faces or (defined? e.vertices and e.vertices.length >= 16)
				end
		end
	end
	
	#-------------------------------------
	
	dlg.show
end

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

def choose_file dlg
	file = dlg.get_element_value 'filename'
	path = ''
	if file.empty?
		file = Sketchup.active_model.title
		file = 'untitled' if file.empty?
		file += '.t3d'
	else
		file = (file.reverse.split /[\/\\]/, 2).map {|x| x.reverse}
		path = file[1] if file.size > 1
		file = file[0]
	end
	
	file = UI.savepanel 'Export to...', path, file
	dlg.execute_script 'select_file(' + file.dump + ')' if file
end

#------------------------------------------------------------------------------#

# Keeps track of the progress during export and shows it to the user

class Progress

attr_reader :error

def initialize dlg
	@dlg = dlg
	@error = false
	
	@stack = []
	
	progress 0
	status ''
end

def progress p = nil
	p = percentage if p.nil?
	@dlg.execute_script 'progress(' + (Float p).to_s + ')'
end

def status msg
	@dlg.execute_script 'status(' + msg.dump + ')' unless @error
end

def error msg
	@dlg.execute_script 'error(' + msg.dump + ')'
	@error = true
end

def percentage
	return 100 if @stack.empty?
	
	p = 0
	f = 100
	
	@stack.each do |x|
		p += f * (Float(x[0]) / Float(x[1]))
		f *= (1.0 / x[1])
	end
	
	p
end

def push count
	@stack << [0,count]
end

def pop
	return if @stack.empty?
	
	@stack.last[0] += 1
	return if @stack.last[0] < @stack.last[1]
	
	@stack.pop
	#pop
end

def abort state
	@dlg.execute_script 'cancel_button(' + state.to_s + ')'
end

end # class Progress

#------------------------------------------------------------------------------#

# Export current model as file (selected model when there is anything selected)

def export filename
	raise 'No file specified.' if filename.empty?
	
	map = T3D::Map.new
	
	model = Sketchup.active_model
	groups = model.layers.map {|x| x = (x.name == 'Layer0' ? 'None' : x.name)}.join(',')
	model = model.selection.empty? ? model.entities : model.selection
	
	map.name 'Model'
	map.name 'Model'
	
	map << T3D::Actor.new('Class' => 'LevelInfo','Name' => map.name('LevelInfo'))
	export_leaf map
	map << T3D::Property.new('VisibleGroups', groups.dump)
	
	@progress.abort true
	@progress.status 'Exporting...'
	export_model map, model
	@progress.pop
	
	@progress.status 'Writing to file...'
	@progress.push map.lines
	
	file = File.new filename, 'w'
	map.write {|x| file.write x; @progress.pop}
	file.close
	
	@progress.pop
	@progress.progress 100
	@progress.status 'Success!'
	@progress.abort false
end

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

def export_leaf map
	levelinfo = map.actors.first
	map << T3D::Property.new('Level', levelinfo.to_s);
	map << T3D::Property.new('Region', 'Zone' => levelinfo.to_s, 'iLeaf' => -1);
end

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

def export_model map, model, trans = Geom::Transformation.new, root = true
	brush = nil
	group = nil
	
	brush = export_brush map, true if root
	group = export_group map, model, true if root
	
	@progress.push model.count
	
	model.each do |e|
		
		case e
		
			when Sketchup::Group
				export_model map, e.entities, trans * e.transformation, false
			
			when Sketchup::ComponentInstance
				export_model map, e.definition.entities, trans * e.transformation, false
			
			when Sketchup::Face
				brush = export_brush(map, root) unless brush
				group = export_group(map, model, root) unless group
				export_face brush, e, trans
		
		end
		
		@progress.pop
	
	end
	
	export_builder_brush brush if brush.polylist.polys.empty? unless brush.nil?
end

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

def export_brush map, root
	map << T3D::Actor.new('Class' => 'Brush', 'Name' => map.name('Brush'))
	map << T3D::Brush.new('Name' => (root ? 'Brush' : map.name('Model')))
	map << T3D::PolyList.new
	
	brush = map.actors.last.brush
	
	export_leaf map
	map << T3D::Property.new('Brush', brush.to_s)
	map << T3D::Property.new('bSelected', 'True') if root
	
	brush
end

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

def export_group map, model, root
	group = root ? 'Cube' : 'None'
	
	if defined? model.parent
		model = model.parent
		model = model.instances[0] if model.group? if defined? model.group?
		name = model.layer.name
		group = (name + ',None') unless root or name == 'Layer0'
	end
	
	map << T3D::Property.new('Group', group)
end

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

def export_builder_brush brush
	cube = [[[0,0,0],[0,0,1],[0,1,1],[0,1,0]],
          [[0,1,0],[0,1,1],[1,1,1],[1,1,0]],
          [[1,1,0],[1,1,1],[1,0,1],[1,0,0]],
          [[1,0,0],[1,0,1],[0,0,1],[0,0,0]],
          [[0,1,1],[0,0,1],[1,0,1],[1,1,1]],
          [[0,0,0],[0,1,0],[1,1,0],[1,0,0]]]
	
	cube.each do |face|
		brush << T3D::Polygon.new
		face.each do |vertex|
			brush << T3D::Coord.new('Vertex', vertex.map {|c| (c * 256) - 128})
		end
	end
	
	brush
end

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

def export_face brush, face, trans
	brush << T3D::Polygon.new
	
	@progress.error 'Warning: some of the faces have more then 16 vertices. ' +
	                'This will crash ued2 on build!!' if face.vertices.length >= 16
	
	face.vertices.each do |v|
		p = trans * v.position
		brush << (T3D::Coord.new 'Vertex', p.to_a.map {|x| x / 1.m})
	end
end

#------------------------------------------------------------------------------#

def show_culprits model
	found = false
	model.each do |e|
		inside = false
		inside = show_culprits (e.entities) {|x| yield x} if defined? e.entities
		inside = show_culprits (e.definition.entities) {|x| yield x} if defined? e.definition.entities
		
		if yield e or inside
			e.hidden = false
			found = true
		else
			e.hidden = true
		end
	end
	found
end

#..............................................................................#