#
# T3D exporter for Google SketchUp
# t3d.rb -- t3d file abstraction
# Created by Ferry Timmers
# June 2010
#

#------------------------------------------------------------------------------#

class T3D

#------------------------------------------------------------------------------#

class Map

attr_accessor :actors

def initialize
	@actors = []
	@count = {}
end

def << actor
	if actor.instance_of? Actor
		@actors << actor
	elsif @actors.last
		@actors.last << actor
	end
end

def lines
	c = 2
	@actors.each {|a| c += a.lines}
	c
end

def name classname
		if @count[classname]
			name = classname + @count[classname].to_s
			@count[classname] += 1
		else
			name = classname + '0'
			@count[classname] = 1
		end
		
		name
end

def write
	yield 'Begin Map'+$/
	@actors.each {|x| x.write {|y| yield y }}
	yield 'End Map'+$/
end

def read
	@actors = []
	
	x = yield
	
	raise 'Parse error' if !x['Begin Map']
	
	while x = yield
		return if x['End Map']
		
		if x['Begin Actor']
			a = Actor.new
			a.read { r = x
				       x = yield
				       r         }
			@actors << a
			
			redo if x
			break
		end
	end
	
	raise 'Parse error'
end
 
end # class Map

#------------------------------------------------------------------------------#

class Actor

attr_accessor :atts, :props, :brush

def initialize args = nil
	args = {} if !args
	raise 'Invalid argument' if not args.kind_of? Hash
	
	@atts = args
	@props = []
	@brush = nil
	
	props << Property.new('Tag', @atts['Class']) if @atts['Class']
	props << Property.new('Name', @atts['Name']) if @atts['Name']
end

def << entity
	if entity.instance_of? Property
		@props << entity
	elsif entity.instance_of? Brush
		@brush = entity
	elsif @brush
		@brush << entity
	end
end

def to_s
	raise 'Actor is undefined' if !@atts['Class'] or !@atts['Name']
	@atts['Class'] + "'MyLevel." + @atts['Name'] + "'"
end

def lines
	c = 2 + @props.length
	c += @brush.lines unless @brush.nil?
	c
end

def write
	atts = (@atts.to_a.map {|a| ' ' + a[0].to_s + '=' + a[1].to_s}).join
	yield 'Begin Actor'+atts+$/
	@brush.write {|x| yield ' '*4 + x} if @brush
	@props.each {|x| x.write {|y| yield ' '*4 + y }}
	yield 'End Actor'+$/
end

def read
	@props = []
	@brush = nil
	
	x = yield
	
	raise 'Parse error' if !x['Begin Actor']
	
	x['Begin Actor'] = ''
	@atts = Hash[*(x.split.map {|p| p.split '='}).flatten]
	
	while x = yield
		return if x['End Actor']
		
		if x['Begin Brush']
			b = Brush.new
			b.read { r = x
				       x = yield
				       r         }
			@brush = b
			
			redo if x
			break
		else
			p = Property.new '', ''
			p.read { r = x
				       x = yield
				       r         }
			@props << p
			
			redo if x
			break
		end
	end
	
	raise 'Parse error'
end

end # class Actor

#------------------------------------------------------------------------------#

class Brush

attr_accessor :atts, :polylist

def initialize args = nil
	args = {} if !args
	raise 'Invalid argument' if not args.kind_of? Hash
	@atts = args
	@polylist = nil
end

def << polylist
	if polylist.instance_of? PolyList
		@polylist = polylist
	elsif @polylist
		@polylist << polylist
	end
end

def to_s
	raise 'Brush is undefined' if !@atts['Name']
	"Model'MyLevel." + @atts['Name'] + "'"
end

def lines
	c = 2
	c += @polylist.lines unless @polylist.nil?
	c
end

def write
	atts = (@atts.to_a.map {|a| ' ' + a[0].to_s + '=' + a[1].to_s}).join
	yield 'Begin Brush'+atts+$/
	@polylist.write {|x| yield ' '*4 + x} if @polylist
	yield 'End Brush'+$/
end

def read
	@polylist = nil
	
	x = yield
	
	raise 'Parse error' if !x['Begin Brush']
	
	x['Begin Brush'] = ''
	@atts = Hash[*(x.split.map {|p| p.split '='}).flatten]
	
	while x = yield
		return if x['End Brush']
		
		if x['Begin PolyList']
			p = PolyList.new
			p.read { r = x
				       x = yield
				       r         }
			@polylist = p
			
			redo if x
			break
		end
	end
	
	raise 'Parse error'
end

end # class Brush

#------------------------------------------------------------------------------#

class PolyList

attr_accessor :atts, :polys

def initialize args = nil
	args = {} if !args
	raise 'Invalid argument' if not args.kind_of? Hash
	@atts = args
	@polys = []
end

def << poly
	if poly.instance_of? Polygon
		@polys << poly
	elsif @polys.last
		@polys.last << poly
	end
end

def lines
	c = 2
	@polys.each {|p| c += p.lines}
	c
end

def write
	atts = (@atts.to_a.map {|a| ' ' + a[0].to_s + '=' + a[1].to_s}).join
	yield 'Begin PolyList'+atts+$/
	@polys.each {|x| x.write {|y| yield ' '*4 + y}}
	yield 'End PolyList'+$/
end

def read
	@polys = []
	
	x = yield
	
	raise 'Parse error' if !x['Begin PolyList']
	
	x['Begin PolyList'] = ''
	@atts = Hash[*(x.split.map {|p| p.split '='}).flatten]
	
	while x = yield
		return if x['End PolyList']
		
		if x['Begin Polygon']
			p = Polygon.new
			p.read { r = x
				       x = yield
				       r         }
			@polys << p
			
			redo if x
			break
		end
	end
	
	raise 'Parse error'
end

end # class PolyList

#------------------------------------------------------------------------------#

class Polygon

attr_accessor :atts, :coords

def initialize args = nil
	args = {} if !args
	raise 'Invalid argument' if not args.kind_of? Hash
	@atts = args
	@coords = []
end

def << coord
	if coord.instance_of? Coord
		@coords << coord
	end
end

def lines
	2 + @coords.length
end

def write
	atts = (@atts.to_a.map {|a| ' ' + a[0].to_s + '=' + a[1].to_s}).join
	yield 'Begin Polygon'+atts+$/
	@coords.each {|x| x.write {|y| yield ' '*4 + y}}
	yield 'End Polygon'+$/
end

def read
	@coords = []
	
	x = yield
	
	raise 'Parse error' if !x['Begin Polygon']
	
	x['Begin Polygon'] = ''
	@atts = Hash[*(x.split.map {|p| p.split '='}).flatten]
	
	while x = yield
		return if x['End Polygon']
		
		c = Coord.new '', []
		c.read { r = x
			       x = yield
			       r         }
		@coords << c
		
		redo if x
		break
	end
	
	raise 'Parse error'
end

end # class Polygon

#------------------------------------------------------------------------------#

class Property

attr_accessor :key, :value

def initialize key, value
	@key = key
	@value = value
end

def write
	yield @key.to_s + '=' + ((@value.kind_of? Hash) ?
		'(' + (@value.to_a.map {|a| a[0].to_s + '=' + a[1].to_s}.join ',') + ')' :
		@value.to_s ) + $/
end

def read
	@key = ''
	@value = ''
	
	x = yield.to_s.strip.split '=', 2
	
	raise 'Parse error' if x.length < 2
	
	@key = x[0]
	@value = (x[1][0] == '(') ?
		Hash[*(x[1][1...-1].split ',').map {|p| p.split '='}.flatten] :
		x[1]
end

end # class Property

#------------------------------------------------------------------------------#

class Coord

attr_accessor :key, :values

def initialize key, values
	raise 'Invalid argument' if not values.kind_of? Array
	@key = key
	@values = values
end

def write
	yield (@key.ljust 9) +
		(@values.map {|x| ('%013.6f' % x).gsub /^0/, '+'}.join ',') + $/
end

def read
	@key = ''
	@values = []
	
	x = yield.strip.split ' '
	
	raise 'Parse error' if x.length != 2
	
	@key = x[0]
	@values = (x[1].split ',').map {|p| Float p}
end

end # class Coord

#------------------------------------------------------------------------------#

end # class T3D

#..............................................................................#